const NOS = require('../submodule/nos-nodejs-sdk/index.js')
const parseXml = require('xml2js').parseString
const fs = require('fs-extra-promise')
const Promise = require('bluebird')
const base = require('zoro-base')
const logger = require('../lib/logger')
// const crypto = require('crypto')

const ResTypeXml = 'xml'
const MaxFileSize = 100 * 1024 * 1024
const MinSliceSize = 16 * 1024

function NOSAsync (options) {
  this.setHost(options.host)
  this.setPort(options.port)
  this.setAccessId(options.accessId)
  this.setSecretKey(options.secretKey)
  this.bucket = options.bucket
}

const proto = NOSAsync.prototype = Object.create(NOS.prototype)

proto.cb = function (options, res) {
  const {resolve, reject, resType} = options
  // 处理 headers
  const headers = res.headers
  var nameHeader = 'x-nos-object-name'
  if (headers[nameHeader]) {
    headers[nameHeader] = decodeURIComponent(headers[nameHeader])
  }

  res.on('error', Error => {
    reject({Error, headers})
  })

  // 接收数据
  res.setEncoding('utf8')
  let body = ''
  res.on('data', chunk => {
    body += chunk
  })

  // 处理结果
  res.on('end', () => {
    if (resType === ResTypeXml && body) {
      parseXml(body, (Error, obj) => {
        if (Error) {
          reject({Error, headers})
        } else {
          obj = obj || {}
          obj.headers = headers
          if (obj.Error) {
            reject(obj)
          } else {
            resolve(obj)
          }
        }
      })
    } else if (resType === 'file') {
      // const md5 = crypto.createHash('md5').update(body).digest('hex')
      const {filePath} = options
      if (filePath) {
        fs.writeFileAsync(filePath, body).then(() => {
          resolve({headers})
        }, Error => {
          reject({Error, headers})
        })
      } else {
        resolve({body, headers})
      }
    } else {
      resolve({headers})
    }
  })
}

// 上传一个对象
proto.putObjectAsync = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.put_object(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 拷贝一个对象
proto.copyObjectAsync = function (options) {
  base.fillUndef(options, {
    src_bucket: this.bucket,
    dest_bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.copy_object(options, this.cb.bind(this, {
      resolve,
      reject
    }))
  })
}

// 桶内部move一个对象
proto.moveObjectAsync = function (options) {
  base.fillUndef(options, {
    src_bucket: this.bucket,
    dest_bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.move_object(options, this.cb.bind(this, {
      resolve,
      reject
    }))
  })
}

// 获取对象相关元数据信息
proto.headObjectAsync = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.head_object(options, this.cb.bind(this, {
      resolve,
      reject
    }))
  })
}

// 读取对象内容
proto.getObjectAsync = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.get_object(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: 'file',
      filePath: options.filePath
    }))
  })
}

// 获取一个桶的对象列表
proto.listObjectsAsync = function (options) {
  options = options || {}
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.list_objects(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 删除一个对象
proto.deleteObjectAsync = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.delete_object(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 删除同一个Bucket中的多个对象
proto.deleteObjectsAsync = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  options.quiet = options.quiet ? 'True' : 'False'
  options.keys = options.keys.map(key => {
    return {Key: key}
  })
  return new Promise((resolve, reject) => {
    this.delete_objects(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 初始化分块上传
proto.createMultipartUploadAsync = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.create_multipart_upload(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 上传一个分块
proto.uploadPart = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.upload_part(options, this.cb.bind(this, {
      resolve,
      reject
    }))
  })
}

// 完成分块上传
proto.completeMultipartUpload = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  options.info = options.etags.map((etag, index) => {
    return {
      PartNumber: index + 1,
      ETag: etag
    }
  })
  return new Promise((resolve, reject) => {
    this.complete_multipart_upload(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 取消分块上传并删除已上传的分块
proto.abortMultipartUpload = function (options) {
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.abort_multipart_upload(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 列出所有执行中的分块上传事件
proto.listMultipartUpload = function (options) {
  options = options || {}
  base.fillUndef(options, {
    bucket: this.bucket
  })
  return new Promise((resolve, reject) => {
    this.list_multipart_upload(options, this.cb.bind(this, {
      resolve,
      reject,
      resType: ResTypeXml
    }))
  })
}

// 取消所有执行中的分块上传事件并删除已上传的分块
proto.clearMultipartUpload = function () {
  return this.listMultipartUpload().then(obj => {
    return Promise.map(obj.ListMultipartUploadsResult.Upload || [], item => {
      return this.abortMultipartUpload({
        key: item.Key[0],
        upload_id: item.UploadId[0]
      })
    })
  })
}

proto.putFilePathAsync = function (options) {
  const {key, filePath} = options
  return fs.statAsync(filePath).then(stats => {
    if (stats.isDirectory()) {
      logger.log('hit directory %s', filePath)
    } else if (stats.isFile()) {
      logger.info('uploading %s to %s', filePath, key)
      return fs.readFileAsync(filePath)
    }
  }).then(buffer => {
    if (!buffer) {
      return
    }
    const bufferSize = buffer.length
    const shouldSlice = bufferSize > MaxFileSize
    // 是否因该切片
    if (shouldSlice || options.forceSlice) {
      return this.createMultipartUploadAsync(options).then(obj => {
        const uploadId = obj.InitiateMultipartUploadResult.UploadId[0]
        // 计算切片大小并切片
        const sliceSize = shouldSlice ? MaxFileSize : MinSliceSize
        const slices = []
        let cursor = 0
        while ((bufferSize - cursor) >= sliceSize * 2) {
          slices.push(buffer.slice(cursor, cursor + sliceSize))
          cursor += sliceSize
        }
        slices.push(buffer.slice(cursor))
        // 上传切片
        return Promise.reduce(slices, (etags, body, index) => {
          return this.uploadPart({
            key,
            part_num: index + 1,
            upload_id: uploadId,
            body
          }).then(obj => {
            etags.push(obj.headers.etag)
            return etags
          })
        }, []).then(etags => {
          return this.completeMultipartUpload({
            key,
            upload_id: uploadId,
            etags
          })
        })
      })
    } else {
      options.body = buffer
      return this.putObjectAsync(options)
    }
  })
}

module.exports = NOSAsync
